(* File: mirror.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

open Common
open Point

type mirror = float * float * float

let make_normal a b c =
  if (a =~ 0. && b =~ 0. && c =~ 0.) then failwith "a = b = c = 0";
  let square_vnorm = a *. a +. b *. b +. c *. c in
  if square_vnorm =~ 1. then (a, b, c)
  else
    let vnorm = sqrt square_vnorm in
    (a /. vnorm, b /. vnorm, c /. vnorm)

let rec make_points p1 p2 = match (p1, p2) with
  | Cartesian (x, y, z), Cartesian (x', y', z') -> make_normal
    (y *. z' -. z *. y') (z *. x' -. x *. z') (x *. y' -. y *. x')
  | p1, p2 -> make_points (to_cartesian p1) (to_cartesian p2)

let rec eval (a, b, c) = function
  | (Spherical _) as p -> eval (a, b, c) (to_cartesian p)
  | Cartesian (x, y, z) -> a *. x +. b *. y +. c *. z

let distance m p = abs_float (eval m p)

let rec image m = function
  | (Spherical _) as p -> image m (to_cartesian p)
  | (Cartesian (x, y, z)) ->
    let a, b, c = m in
    let factor = 2. *. (a *. x +. b *. y +. c *. z) in
    Cartesian (x -. a *. factor, y -. b *. factor, z -. c *. factor)
