(* File: point.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

open Common

type point =
  | Spherical of (float * float)
  | Cartesian of (float * float * float)

let to_cartesian = function
  | Spherical (phi, theta) ->
    Cartesian (sin phi *. cos theta, sin phi *. sin theta, cos phi)
  | p -> p

let to_spherical = function
  | Cartesian (x, y, z) -> Spherical (acos z, atan (y /. x))
  | p -> p

let rec ( =* ) p p' = match (p, p') with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    x =~ x' && y =~ y' && z =~ z'
  | Spherical (a, b), Spherical (a', b') ->
    a =~ a' && b =~ b'
  | p, p' -> to_cartesian p =* to_cartesian p'

let rec distance p p' = match (p, p') with
  | Cartesian (x, y, z), Cartesian (x', y', z') ->
    let dx = x -. x' and dy = y -. y' and dz = z -. z' in
    sqrt (dx *. dx +. dy *. dy +. dz *. dz)
  | p, p' -> distance (to_cartesian p) (to_cartesian p')

let rec to_string ?(after="") = function
  | Cartesian (x, y, z) -> Printf.sprintf "(%.10f, %.10f, %.10f)%s" x y z after
  | (Spherical _) as p -> to_string (to_cartesian p)

let center points =
  let sx = ref 0. and sy = ref 0. and sz = ref 0. in
  let factor = 1. /. float (Array.length points) in
  Array.iter (function
  | Cartesian (x, y, z) -> sx := !sx +. x; sy := !sy +. y; sz := !sz +. z
  | _ -> assert false) (Array.map to_cartesian points);
  Cartesian (!sx *. factor, !sy *. factor, !sz *. factor)

