(* File: wythoff.ml

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

open Common
open Printf
open Scanf

module M = Mirror
module P = Point

let ( =* ) = P.( =* )

(* from which border does the triangle come from ? *)
type from = P | Q | R | Start
type triangle = { p: P.point; q: P.point; r: P.point; x: P.point;
                  from: from; xid: int }

type start = One | Bissected | Incenter | Snub | Custom of P.point

type polyhedron = {
  vertices: Point.point array;
  (** Edges are pairs (a, b) of natural integers indexing a vertex in the
      vertices array, where a must be smaller than b *)
  edges: (int * int) list;
  (** Faces are lists [i1, i2, ..., iN] of vertices indices, (iX, i(X+1))
      must form an existing edge (or its inverse (i(X+1), iX)), as must be
      (i1, iN) *)
  faces: int list list
}

let triangle_to_string t =
  let s = P.to_string in
  sprintf "(%s; %s; %s; %s (%s))\n%!" (s t.p) (s t.q) (s t.r) (s t.x)
    (match t.from with | P -> "P" | Q -> "Q" | R -> "R" | Start -> "Start")

let triangle_is_done triangles t =
  let ( =* ) = P.( =* ) in
  let is_equal x = x.p =* t.p && x.q =* t.q && x.r =* t.r in
  let equals = List.filter is_equal triangles in
  List.length equals > 0

let getmirror mirrors p1 p2 =
  let candidates = List.filter (fun x -> M.distance x p1 =~ 0.) !mirrors in
  let candidates = List.filter (fun x -> M.distance x p2 =~ 0.) candidates in
  if List.length candidates = 0 then begin
    mirrors := M.make_points p1 p2 :: !mirrors;
    List.hd !mirrors
  end else List.hd candidates

let basetriangle pn pd qn qd rn rd barpos =
  let p = pi *. float pd /. float pn in
  let q = pi *. float qd /. float qn in
  let r = pi *. float rd /. float rn in
  let cosp = cos p and cosq = cos q and cosr = cos r in
  let sinp = sin p and sinq = sin q and sinr = sin r in
  let p0q = acos ((cosr +. cosp *. cosq) /. (sinp *. sinq)) in
  let p0r = acos ((cosq +. cosp *. cosr) /. (sinp *. sinr)) in
  let q0r = acos ((cosp +. cosq *. cosr) /. (sinq *. sinr)) in
  let point_p = P.Spherical (0., 0.)
  and point_q = P.Spherical (p0q, 0.)
  and point_r = P.Spherical (p0r, p) in
  let x = match barpos with
    | One -> point_p
    | Bissected -> P.Spherical (p0q /. ( 1. +. q0r /. p0r), 0.)
    | Incenter -> failwith "Not yet implemented"
    | Snub -> failwith "Not yet implemented"
    | Custom x -> x
  in
  {xid=0; p=point_p; q=point_q; r=point_r; from=Start; x=x}

let startingpoint pn pd qn qd rn rd barpos =
  (basetriangle pn pd qn qd rn rd barpos).x

let getxid xvalues x =
  let rec aux = function
    | hd :: tl when hd =* x -> List.length tl
    | hd :: tl -> aux tl
    | [] -> xvalues := x :: !xvalues; pred (List.length !xvalues)
  in aux !xvalues

let addtriangles links xvalues triangles_left mirrors t =
  let addt p q r from mirrorx =
    let x = M.image mirrorx t.x in
    let xid = getxid xvalues x in
    let t' = {p = p; q = q; r = r; from = from; x = x; xid = xid} in
    if xid <> t.xid then links := (min t.xid xid, max t.xid xid) :: !links;
    Queue.push t' triangles_left
  in
  if t.from <> P then begin
    let imp = M.image (getmirror mirrors t.q t.r) t.p in
    addt imp t.q t.r P (getmirror mirrors t.q t.r)
  end;
  if t.from <> Q then begin
    let imq = M.image (getmirror mirrors t.p t.r) t.q in
    addt t.p imq t.r Q (getmirror mirrors t.p t.r)
  end;
  if t.from <> R then begin
    let imr = M.image (getmirror mirrors t.p t.q) t.r in
    addt t.p t.q imr R (getmirror mirrors t.p t.q)
  end

let recup_corners triangles tripoint =
  let f t = match tripoint with
    | (P, p) -> t.p =* p
    | (Q, q) -> t.q =* q
    | (R, r) -> t.r =* r
    | (Start, _) -> assert false
  in List.map (fun t -> t.xid) (List.filter f triangles)

let order_corners edges corners =
  if List.length corners < 4 then corners
  else
    let interresting_edges = List.filter
      (fun (a, b) ->
        List.exists (fun x -> x = a) corners &&
          List.exists (fun x -> x = b) corners) edges
    in
    let ordered = Array.make (List.length corners) (-1) in
    ordered.(0) <- fst (List.hd interresting_edges);
    ordered.(1) <- snd (List.hd interresting_edges);
    for i = 2 to List.length corners - 1 do
      let prec = ordered.(i - 2) in
      let cur = ordered.(pred i) in
      let x, y = List.find
        (fun (a, b) -> a = cur && b <> prec || b = cur && a <> prec)
        interresting_edges
      in
      ordered.(i) <- x + y - cur
    done;
    Array.to_list ordered

let poly pn pd qn qd rn rd barpos =
  let links = ref [] in
  let mirrors = ref [] in
  let basetriangle = basetriangle pn pd qn qd rn rd barpos in
  let xvalues = ref [basetriangle.x] in
  let triangles_done = ref [] in
  (* The use of a queue allows to visit triangles from the most pertinent
     to the less because the first visited are issued from a less number
     of reflections *)
  let triangles_left = Queue.create () in
  let tripoints = ref [] in
  Queue.add basetriangle triangles_left;
  while not (Queue.is_empty triangles_left) do
    let t = Queue.pop triangles_left in
    tripoints := (P, t.p) :: (Q, t.q) :: (R, t.r) :: !tripoints;
    if not (triangle_is_done !triangles_done t) then begin
      triangles_done := t :: !triangles_done;
      addtriangles links xvalues triangles_left mirrors t
    end;
  done;
  let links = unicize !links in
  let comp (from, p) (from', p') = from = from' && p =* p' in
  let tripoints = unicize ~f:comp !tripoints in
  let faces_corners = List.map (recup_corners !triangles_done) tripoints in
  let faces_corners = List.map unicize faces_corners in
  let faces_corners = List.filter (fun l -> List.length l > 2) faces_corners in
  let faces = List.map (order_corners links) faces_corners in
  {vertices = Array.of_list (List.rev !xvalues); edges = links; faces = faces}

let params_from_string s =
  let chars_of_string s =
    let acc = ref [] in
    String.iter (fun x -> acc := x :: !acc) s;
    List.rev !acc
  in
  let barpos = ref None in
  let rec aux ?(fracpart=false) acc = function
    | [] -> (match acc, !barpos with
      | ((rn, rd) :: (qn, qd) :: (pn, pd) :: []), Some bp ->
        pn, pd, qn, qd, rn, rd, bp
      | _ -> assert false)
    | ' ' :: tl | '\n' :: tl | '\t' :: tl | '\r' :: tl -> aux ~fracpart acc tl
    | (('2' .. '5') as hd) :: tl ->
      let x = Char.code hd - Char.code '0' in
      if fracpart then
        aux ((fst (List.hd acc), x) :: (List.tl acc)) tl
      else
        aux ((x, 1) :: acc) tl
    | '/' :: tl -> aux ~fracpart:true acc tl
    | '|' :: tl ->
      assert (not fracpart);
      barpos := Some (match List.length acc with
      | 0 -> Snub
      | 1 -> One
      | 2 -> Bissected
      | 3 -> Incenter
      | _ -> assert false);
      aux acc tl
    | ('(' :: _) as s ->
      sscanf (List.fold_left ( ^ ) "" (List.map (String.make 1) s)) "(%f, %f)"
        (fun x y -> barpos := Some (Custom (P.Spherical (x, y))));
      aux acc []
    | _ -> assert false
  in
  aux [] (chars_of_string s)

let poly_from_string s =
  let pn, pd, qn, qd, rn, rd, barpos = params_from_string s in
  poly pn pd qn qd rn rd barpos
