(* File: wythoff.mli

   Copyright (C) 2011-2015

     Fabian Pijcke <fabian.pijcke@gmail.com>
     Pierre Hauweele <antegallya@gmail.com>

   This library is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License version 3 or
   later as published by the Free Software Foundation, with the special
   exception on linking described in the file LICENSE.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the file
   LICENSE for more details. *)

type start =
  (** The first vertex is on the point P of the PQR. The Wythoff symbol is
      "p | q r" *)
  | One
  (** The first vertex is on the intersection between PQ and the bissector
      of R. Wythoff symbol: "p q | r" *)
  | Bissected
  (** The first vertex is at the incenter of PQR. The resulting Wythoff
      symbol is "p q r |" *)
  | Incenter
  (** The first vertex is placed such that a snub polyhedron is
      generated. Wythoff symbol: "| p q r" *)
  | Snub
  (** The first vertex is placed at the specified point (No Wythoff symbol
      for this case) *)
  | Custom of Point.point

type triangle

type polyhedron = {
  vertices: Point.point array;
  edges: (int * int) list;
  faces: int list list
}

val triangle_to_string : triangle -> string
(** String representation of a triangle *)

val basetriangle : int -> int -> int -> int -> int -> int -> start -> triangle
(** [basetriangle pn pd qn qd rn rd startpoint] returns the first schwartz
    triangle that will be used to generate the polyhedra *)

val startingpoint : int -> int -> int -> int -> int -> int -> start ->
  Point.point
(** [startingpoint pn pd qn qd rn rd startpoint] returns the coordinates
    of the starting point *)

val poly : int -> int -> int -> int -> int -> int -> start -> polyhedron
(** [poly pn pd qn qd rn rd startpoint] returns a triple (vertices, edges,
    faces) which represent the polyhedron with Schwartz triangle ([pn]/[pd]
    [qn]/[qd] [rn]/[rd]), starting point depending on the value of
    [startpoint] *)

val params_from_string : string -> int * int * int * int * int * int * start
(** Get the Wythoff parameters from a Wythoff symbol *)

val poly_from_string : string -> polyhedron
(** Get a polyhedron from a Wythoff symbol *)
